//*****************************************************************************
//*****************************************************************************
//  FILENAME:   sensor.c
//  Version: 1.0, 2007/03/08
//  Generated by Andy Kiser
//
//  DESCRIPTION:  sensor specific functions 
//
//-----------------------------------------------------------------------------
// Originally authored by Toradex AG. This work is now available under the terms and conditions
// of the Creative Commons License 'Attribution CC BY', details of which can be found here:
// http://creativecommons.org/licenses/by/3.0/
//*****************************************************************************
//*****************************************************************************
//#define __DEBUG__
#include <m8c.h>        // part specific constants and macros
#include "sensor.h"
#include "psocgpioint.h"
#include "pwm8.h"
#include "feature.h"
#include "parameters.h"
#include "types_n_macros.h"
//----------------------------------------
// Sensor parameters
// Factory Defaults are defined here (NUM_xxx are defined in "sensor.h")
//----------------------------------------
const BYTE FactoryDefault_Param_8   [NUM_PARAM_8] =
{
	0x00,		// report mode
	0x00		// LED mode
};

const WORD FactoryDefault_Param_16  [NUM_PARAM_16] = 
{
	500,		// report rate [ms]
	500,		// sample rate [ms]
	50000		// sample rate [us]	(not used for this sensor)
};

const DWORD FactoryDefault_Param_32  [NUM_PARAM_32] =
{
	0x00000000	// currently no 32bit parameters
};

const tParam_168 FactoryDefault_Param_168 [NUM_PARAM_168] =
{
	"Pressure Sensor",	// user device name
	"Frame Number",		// user name Channel 0
	"Pressure",			// user name Channel 1
	"Temperature"		// user name Channel 2
};


//----------------------------------------
// this module's global variables
//----------------------------------------
tReport Report;

//----------------------------------------
// constant lookup-table for conversion ADC->physical value
//----------------------------------------


//----------------------------------------
// type definitions
//----------------------------------------

typedef enum 
{	
	eIdle,
	eMeasuring_Temp,
	eMeasuring_Pressure,
} tFSM_Measure;

//----------------------------------------
// variables

tFSM_Measure	FSM;
WORD			Countdown_1ms;

BOOL			NewSample = FALSE;

static	WORD C1;
static	WORD C2_x4;
static	WORD C3;
//static	WORD C4_m512;
static	signed C4_m512;
static	WORD C5_x8;
static	WORD C6;

static	WORD UT1;
static	WORD C6_p50;
static	WORD C6_p24_x11;
static	WORD C6_p24_x3;

//static	WORD OFF, SENS;
static	signed OFF;
static	WORD SENS;


// store original drivemodes during suspend
BYTE			DM0_2, DM0_1, DM0_0;
BYTE			DM2_2, DM2_1, DM2_0;

//----------------------------------------
// function prototypes
//
static void MS5534_ReadCalibration(void);
static void MS5534_Reset(void);

//-----------------------------------------------------------------------------
// Initializes the Sensor 
//-----------------------------------------------------------------------------
void Sensor_Start ()
{
	PWM8_Start();
	MS5534_Reset();
	MS5534_ReadCalibration();
	FSM = eIdle;
}

//-----------------------------------------------------------------------------
// Prepare Sensor for Suspend
//-----------------------------------------------------------------------------
void Sensor_Suspend()
{
	// store original drivemodes during  suspend
	DM0_2 = PRT0DM2;
	DM0_1 = PRT0DM1;
	DM0_0 = PRT0DM0;
	DM2_2 = PRT2DM2;
	DM2_1 = PRT2DM1;
	DM2_0 = PRT2DM0;
		
	MS5534_Reset();
	PWM8_Stop();

	// set all I/Os to/from the sensor to High-Z Analog (=110), 
	PRT0DM2  = 0xFF;
	PRT0DM1  = 0xFF;
	PRT0DM0  = 0x00;
	PRT2DM2  = 0xFF;
	PRT2DM1  = 0xFF;
	PRT2DM0  = 0x00;
}
//-----------------------------------------------------------------------------
// Resume Sensor from Suspend mode
//-----------------------------------------------------------------------------
void Sensor_Resume()
{
	// restore drivemodes for GPIOs to/from the sensor
	PRT2DM2 = DM2_2;
	PRT2DM1 = DM2_1;
	PRT2DM0 = DM2_0;

	PRT0DM2 = DM0_2;
	PRT0DM1 = DM0_1;
	PRT0DM0 = DM0_0;

	FSM = eIdle;
	PWM8_Start();
	MS5534_Reset();
}

//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// cannot be handled with the standard functionality in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//
//	return Value: number of Parameter bytes (1-31) that are 
//				  returned in the IN report
//-----------------------------------------------------------------------------
BYTE Sensor_Specific_Feature(void)
{
	// currently no special feature reports are handled
	return 0;
}

//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// sets / gets a 8/16/32/168bit parameter in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//-----------------------------------------------------------------------------
void Sensor_Specific_Parameter(BYTE *NumberOfReturnedBytes)
{
	// Oak-P does not have specific parameters
}


//-----------------------------------------------------------------------------
// Transmit and Receive <size> bits to the sensor / from the sensor
// - Data bits 0..<size-1> are transmitted, MSB first
// - <size> must be in the range 0..8
// return value: received bits 
//-----------------------------------------------------------------------------
BYTE MS5534_TxRx (BYTE TxData, BYTE size)
{
	BYTE i;
	BYTE Rx=0;
	
	// shift TxData to the left end of the byte
	for (i=size; i<8; i++)
		TxData <<=1;
	
	while (size != 0)
	{
		// set MOSI output 
		if ( (TxData & 0x80) == 0 )
			MOSI_Data_ADDR &= ~MOSI_MASK;
		else
			MOSI_Data_ADDR |=  MOSI_MASK;
		
		// wait 0.5 cycles (use time to do some processing)
		TxData <<= 1;
		Rx <<= 1;
		
		// generate positive clock edge
		SCLK_Data_ADDR |=  SCLK_MASK;
		
		// wait 0.5 cycles (use time to do some processing)
		/*for(i=0; i<10; i++);*/
		size--;
		
		// generate negative clock edge
		SCLK_Data_ADDR &= ~SCLK_MASK;
		
		// read Rx data bit
		if ( (MISO_Data_ADDR & MISO_MASK) != 0)
			Rx++;
	}	
	
	return Rx;
}

//-----------------------------------------------------------------------------
// Send a reset sequence
//-----------------------------------------------------------------------------
static void MS5534_Reset(void)
{
	MS5534_TxRx(0xAA, 8);
	MS5534_TxRx(0xAA, 8);
	MS5534_TxRx(0x00, 5);
}

//-----------------------------------------------------------------------------
// Send Sequence to Start Pressure Measurement
//-----------------------------------------------------------------------------
static void MS5534_MeasureP_Start(void)
{
	MS5534_TxRx(0xFF, 3);
	MS5534_TxRx(0x0A, 4);
	MS5534_TxRx(0x00, 5);
}

//-----------------------------------------------------------------------------
// Send Sequence to Start Temperature Measurement
//-----------------------------------------------------------------------------
static void MS5534_MeasureT_Start(void)
{
	MS5534_TxRx(0xFF, 3);
	MS5534_TxRx(0x09, 4);
	MS5534_TxRx(0x00, 5);
}
//-----------------------------------------------------------------------------
// return TRUE if ADC conversion has finished and Measurement result is ready 
//-----------------------------------------------------------------------------
static BOOL MS5534_IsMeasurementDone(void)
{
	return (MISO_Data_ADDR & MISO_MASK) == 0;
}

//-----------------------------------------------------------------------------
// Read 16 Bit Measurement result or 16 Bit calibration word
//-----------------------------------------------------------------------------
static WORD MS5534_ReadWord(void)
{
	BYTE LSB, MSB;
	
	MSB = MS5534_TxRx(0x00, 8);
	LSB = MS5534_TxRx(0x00, 8);
	MS5534_TxRx(0x00, 1);
	
	return 256*MSB + LSB;
}

//-----------------------------------------------------------------------------
// Read the calibration data from the sensor into global variables C1..C6
//-----------------------------------------------------------------------------
static void MS5534_ReadCalibration(void)
{
	BYTE Raw[8];
	BYTE b, shift, cmd;
	BYTE MSB, LSB;
	WORD wtmp;
	
	// Read raw calibration data from sensor
	for (b=0; b<8; b++,b++)
	{
		if (b==0)		// produces more efficient code than a switch statement
			cmd = 0x15;
		if (b==2)
			cmd = 0x16;
		if (b==4)
			cmd = 0x19;
		if (b==6)
			cmd = 0x1A;
			
		MS5534_TxRx(0xFF, 3);
		MS5534_TxRx(cmd , 6);
		MS5534_TxRx(0x00, 4);
		
		Raw[b  ] = MS5534_TxRx(0x00, 8); // MSB
		Raw[b+1] = MS5534_TxRx(0x00, 8); // LSB
		MS5534_TxRx(0x00, 1);

	}
		//Raw[0] = 0xb7;	//TestValues
		//Raw[1] = 0x5C;
		//Raw[2] = 0x9d;
		//Raw[3] = 0x19;
		//Raw[4] = 0x62;
		//Raw[5] = 0x54;
		//Raw[6] = 0xb8;
		//Raw[7] = 0x6c;
	// move raw calibration data into global variables C1..C6
	C1		= (Raw[0]*256+ Raw[1]); C1>>=1;
	C2_x4	= (Raw[5]&0x3F)*256; C2_x4+=(Raw[7]&0x3F)<<2;
	C3		= (Raw[6]*256+ Raw[7]); C3>>=6;
	//C4_m512	= (Raw[4]*256+ Raw[5]); C4_m512= (C4_m512>>=6) - 512;
	wtmp	= (Raw[4]*256+ Raw[5]);
	C4_m512= (wtmp>>=6) - 512;
	C5_x8	= (Raw[2]*256+ Raw[3]) & 0xFFC0; 
	C5_x8 >>= 3; 
	C5_x8  += ((Raw[1]&0x01)<<5)*256;
	C6		= (Raw[3]&0x3F);

	// precalculate values that are constant for a particular sensor
	UT1 = C5_x8 + 20224;
	C6_p50 = C6 + 50;
	C6_p24_x11 = (C6 + 24)*11;
	C6_p24_x3 = (C6 + 24)*3;
}

//-----------------------------------------------------------------------------
// Turn the raw ADC values into real Pressure and temperature 
// (resolution: 0.1C / 10Pa)
//
// side effects:
// changes global variables
//  - OFF
//  - SENS
//-----------------------------------------------------------------------------
static void MS5534_Compute(WORD D1, WORD D2, WORD *Pressure, WORD *Temp)
{
//	WORD TEMP, P, T2, P2;
//	WORD dT, X;
	signed TEMP, P, T2, P2; 
	signed dT, X;	
//	unsigned long ltmp, sq;
	signed long ltmp, sq; 

		//D1 = 16460; //TestValues
		//D2 = 20000;
	
	// --- basic equations
	dT = D2 - UT1;
	
//	ltmp = (unsigned long)C6_p50 * dT;
//	TEMP = (WORD)((ltmp/256) >>2) + 200; 
	ltmp = (unsigned long)C6_p50 * dT; 
	TEMP = (signed)((ltmp/256) >>2) + 200; 
	
//	ltmp = (unsigned long)C4_m512 * dT;
//	OFF =  (WORD)((ltmp/256)>>4) + C2_x4;
	ltmp = (signed long)C4_m512 * dT;
	OFF =  (signed)((ltmp/256)>>4) + C2_x4;
	
//	ltmp = (unsigned long)C3 * dT;
//	SENS = (WORD)((ltmp/256)>>2) + C1 + 24576;
	ltmp = (unsigned long)C3 * dT;
	SENS = (WORD)((ltmp/256)>>2) + C1 + 24576;
	
//	ltmp = (unsigned long)(D1-7168) * SENS;
//	X 	 = (WORD)((ltmp/256)>>6) - OFF;
	ltmp = (unsigned long)(D1-7168) * SENS;
	X 	 = (signed)((ltmp/256)>>6) - OFF;
	
//	ltmp = (unsigned long)10 * X;
//	P	 = (WORD)(ltmp >>5) + 2500;
	ltmp = (signed long)10 * X;
	P	 = (signed)(ltmp >>5) + 2500;
	
	// --- second order temperature compensation
	if (TEMP < 200)
	{
		sq=200-TEMP;
		ltmp = C6_p24_x11 * sq * sq;
		T2 = (ltmp/65536)>>4;
		
		ltmp = 3 * (signed long)T2 * (P-3500);
		P2 = (ltmp/256)>>6;
	}
	else if (TEMP > 450)	
	{
		sq=TEMP-450;
		ltmp = C6_p24_x3 * sq * sq;
		T2 = (ltmp/65536)>>4;
		
		ltmp = (signed long)T2 * (P-10000);
		P2 = (ltmp/256)>>5;
	}
	else
	{
		T2 = 0;
		P2 = 0;
	}
	
	
	// return results
	T2 = TEMP-T2;	
	P2 = P- P2;	
	*Temp = T2 + 2731;	// +273.1 for conversion degC->K
	*Pressure = P2;
}

//-----------------------------------------------------------------------------
// void Sensor_FSM_Measure(void)
//
//	finite state machine to control the measurement process for 
//  the sensor
//
//  IMPORTANT: 	call this routine once every millisecond. Otherwise the timing
//				parameters cannot be met.
//-----------------------------------------------------------------------------
void Sensor_FSM_Measure(void) 
{
	WORD		Pressure;
	WORD		Temp;
	static WORD Pressure_Raw; // Pressure ADC value
	static WORD Temp_Raw; // Temperature ADC value

	// decrement millisecond counter
	if (Countdown_1ms != 0)
		Countdown_1ms--;

	// process actual state machine
	switch (FSM)
	{
		case eIdle: 
			if (Countdown_1ms == 0)
			{
				Get_Param_16(PAR16_SAMPLERATE_MS, VOLATILE, &Countdown_1ms);;
			
				// start temperature measurement
				MS5534_MeasureT_Start();
		
				FSM = eMeasuring_Temp;
			}
			break;			
			
		case eMeasuring_Temp:
			if (MS5534_IsMeasurementDone())
			{
				Temp_Raw = MS5534_ReadWord();	// read temperature ADC
				
				// start pressure measurement
				MS5534_MeasureP_Start();
				FSM = eMeasuring_Pressure;
			}
			break;			
		
		case eMeasuring_Pressure:
			if (MS5534_IsMeasurementDone())
			{
				Pressure_Raw = MS5534_ReadWord();	// read temperature ADC
				
				// calculate actual Pressure and Temperature (calibrated)
				MS5534_Compute(Pressure_Raw, Temp_Raw, &Pressure, &Temp);

				// fill measured data into IN REPORT
				Report.Ch1_LSB = Pressure & 0xFF;
				Report.Ch1_MSB = Pressure / 0x100;
				Report.Ch2_LSB = Temp & 0xFF;
				Report.Ch2_MSB = Temp / 0x100;

				NewSample=TRUE;
				
				// Reset the sensor
				MS5534_Reset();
				FSM = eIdle;
			}
			break;			
	}
}


//-----------------------------------------------------------------------------
void Sensor_Debug(void)
{
/* currently no debug function */
}